﻿using System;
using System.Collections.Generic;
using System.Data;
using System.Globalization;
using System.IO;
using System.Runtime.Remoting.Messaging;
using log4net.Appender;
using log4net.Core;
using log4net.Util;

namespace AsyncAdoNetLogging
{
    public class AsyncAdoNetAppender : AppenderSkeleton
    {
        private string m_commandText;
        private CommandType m_commandType = CommandType.Text;
        private string m_connectionString;
        private string m_connectionType =
            "System.Data.OleDb.OleDbConnection, System.Data, Version=1.0.3300.0, Culture=neutral, PublicKeyToken=b77a5c561934e089";

        protected List<AdoNetAppenderParameter> m_parameters = new List<AdoNetAppenderParameter>();        
        private SecurityContext m_securityContext;
        protected bool m_usePreparedCommand;
        
        public string CommandText
        {
            get { return m_commandText; }
            set { m_commandText = value; }
        }

        public CommandType CommandType
        {
            get { return m_commandType; }
            set { m_commandType = value; }
        }

        public string ConnectionString
        {
            get { return m_connectionString; }
            set { m_connectionString = value; }
        }

        public string ConnectionType
        {
            get { return m_connectionType; }
            set { m_connectionType = value; }
        }

        public SecurityContext SecurityContext
        {
            get { return m_securityContext; }
            set { m_securityContext = value; }
        }

        public override void ActivateOptions()
        {
            base.ActivateOptions();
            m_usePreparedCommand = !string.IsNullOrEmpty(m_commandText);
            if (m_securityContext == null)
            {
                m_securityContext = SecurityContextProvider.DefaultProvider.CreateSecurityContext(this);
            }
        }

        public void AddParameter(AdoNetAppenderParameter parameter)
        {
            m_parameters.Add(parameter);
        }

        protected virtual string GetLogStatement(LoggingEvent logEvent)
        {
            if (Layout == null)
            {
                ErrorHandler.Error("ADOAppender: No Layout specified.");
                return "";
            }
            var writer = new StringWriter(CultureInfo.InvariantCulture);
            Layout.Format(writer, logEvent);
            return writer.ToString();
        }

        private IDbCommand InitializeDatabaseCommand(IDbConnection connection)
        {
            IDbCommand command = null;
            if ((connection != null) && m_usePreparedCommand)
            {   
                try
                {
                    command = connection.CreateCommand();
                    command.CommandText = m_commandText;
                    command.CommandType = m_commandType;
                }
                catch (Exception exception)
                {
                    ErrorHandler.Error("Could not create database command [" + m_commandText + "]", exception);
                    if (command != null)
                    {
                        try
                        {
                            command.Dispose();
                        }
                        catch
                        {
                        }
                        command = null;
                    }
                }
                if (command != null)
                {
                    try
                    {
                        foreach (AdoNetAppenderParameter parameter in m_parameters)
                        {
                            try
                            {
                                parameter.Prepare(command);
                            }
                            catch (Exception exception)
                            {
                                ErrorHandler.Error(
                                    "Could not add database command parameter [" + parameter.ParameterName + "]",
                                    exception);
                                throw;
                            }
                        }
                    }
                    catch
                    {
                        try
                        {
                            command.Dispose();
                        }
                        catch
                        {
                        }
                        command = null;
                    }
                }
                if (command != null)
                {
                    try
                    {
                        command.Prepare();
                    }
                    catch (Exception exception)
                    {
                        ErrorHandler.Error("Could not prepare database command [" + m_commandText + "]", exception);
                        try
                        {
                            command.Dispose();
                        }
                        catch
                        {
                        }
                        command = null;
                    }
                }
            }
            return command;
        }

        private IDbConnection CreateConnection()
        {
            IDbConnection connection = (IDbConnection)Activator.CreateInstance(ResolveConnectionType());
            connection.ConnectionString = m_connectionString;
            using (SecurityContext.Impersonate(this))
            {
                connection.Open();
            }
            return connection;
        }

        protected virtual Type ResolveConnectionType()
        {
            Type type;
            try
            {
                type = SystemInfo.GetTypeFromString(m_connectionType, true, false);
            }
            catch (Exception exception)
            {
                ErrorHandler.Error("Failed to load connection type [" + m_connectionType + "]", exception);
                throw;
            }
            return type;
        }

        protected override void Append(LoggingEvent evt)
        {
            IDbConnection connection = CreateConnection();
            IDbCommand command = InitializeDatabaseCommand(connection );

            IDbTransaction dbTran = connection.BeginTransaction();
            Append(dbTran, command, evt);
        }

        protected virtual void Append(IDbTransaction dbTran, IDbCommand cmd, LoggingEvent evt)
        {
            if (cmd != null)
            {
                if (dbTran != null)
                {
                    cmd.Transaction = dbTran;
                }

                foreach (AdoNetAppenderParameter parameter in m_parameters)
                {
                    parameter.FormatValue(cmd, evt);
                }

                Func<int> asyncDbCommand = cmd.ExecuteNonQuery;
                asyncDbCommand.BeginInvoke(SqlExecutionCallBack, dbTran);
            }
        }

        private void SqlExecutionCallBack(IAsyncResult ar)
        {
            var dbTransaction = (IDbTransaction) ar.AsyncState;
            try
            {
                var asyncResult = (AsyncResult) ar;
                var asyncDbCommand = (Func<int>) asyncResult.AsyncDelegate;
                asyncDbCommand.EndInvoke(ar);
                dbTransaction.Commit();
            }
            catch (Exception ex)
            {
                dbTransaction.Rollback();
                ErrorHandler.Error("Exception while writing to database", ex);
            }
        }
    }
}